import org.apache.commons.cli.*;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;
import org.bson.*;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.time.Duration;
import java.time.Instant;
import java.util.*;

public class SubscribeFromKafka {
    private static final Logger log = LoggerFactory.getLogger(SubscribeFromKafka.class);
    private static Map<Integer, partitionMsgConsumer> partitionMsgConsumers = new HashMap<>();

    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();

        Option option_isPrintCkp = Option.builder()
                .required(false)
                .desc("print checkpoint or not")
                .longOpt("isPrintCkp")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_isPrintCkp);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("isPrintCkp")
            );

        } catch (Exception exception) {
            log.error(exception.getMessage());
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean isPrintCkp) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, false);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, 20 << 20); // kafka消息上限20M
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArrayDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        KafkaConsumer<String, byte[]> consumer = new KafkaConsumer<>(props);
        consumer.subscribe(Collections.singletonList(topic), new partitionSubLis());
        try {
            while (true) {
                ConsumerRecords<String, byte[]> records = consumer.poll(Duration.ofSeconds(1));
                for (ConsumerRecord<String, byte[]> record : records) {
                    partitionMsgConsumer pc = partitionMsgConsumers.get(record.partition());
                    if (null == pc) {
                        throw new IllegalStateException("no such partition consumer " + record.partition());
                    }

                    pc.consume(consumer, record, isPrintCkp);
                }
            }
        }catch (Exception e ){
            log.error(e.getMessage());
        }
        finally {
            consumer.close();
        }
    }



    private static class partitionMsgConsumer {
        TopicPartition tp;

        private partitionMsgConsumer(TopicPartition tp) {
            this.tp = tp;
        }

        private void consume(KafkaConsumer<String, byte[]> consumer, ConsumerRecord<String, byte[]> record, boolean isPrintCkp) {


            RawBsonDocument rawBsonDocument;

            boolean isCKP = false;

            for (Header header : record.headers()) {
                //提交checkpoint
                if (header.key().equals(Const.CHECKPOINT)) {
                    consumer.commitAsync(Collections.singletonMap(tp, new OffsetAndMetadata(record.offset() + 1)), null);
                    isCKP = true;
                    break;
                }
            }

            if (isCKP) {
                if (isPrintCkp) {
                    log.info("partition:{}, offset:{}, header:{}, value:{}", record.partition(), record.offset(), record.headers(), record.value());
                }
            }else {
                rawBsonDocument  = new RawBsonDocument(record.value());

                log.info("partition:{}, offset:{}, value:{}", record.partition(), record.offset(), rawBsonDocument.toJson());
            }

        }

    }



    private static class partitionSubLis implements ConsumerRebalanceListener {
        public void onPartitionsAssigned(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.computeIfAbsent(p.partition(), k -> new partitionMsgConsumer(p));
            }
        }

        public void onPartitionsRevoked(Collection<TopicPartition> partitions) {
            for (TopicPartition p : partitions) {
                partitionMsgConsumers.remove(p.partition());
            }
        }
    }
}
